"""Error and Warning Reporting Module"""
from os.path import join, abspath, dirname
import copy
from typing import Optional, Dict, List, Tuple
from json import dumps, JSONEncoder
from yaml import load, BaseLoader

from reasoner_validator.versioning import latest


class CodeDictionary:

    CODE_DICTIONARY: str = join(abspath(dirname(__file__)), "codes.yaml")

    code_dictionary: Optional[Dict] = None

    @classmethod
    def _get_code_dictionary(cls) -> Dict:
        if not cls.code_dictionary:
            # Open the file and load the file
            with open(cls.CODE_DICTIONARY, mode='r') as f:
                cls.code_dictionary = load(f, Loader=BaseLoader)
        return cls.code_dictionary

    @classmethod
    def _get_nested_code_value(cls, data: Dict, path: List[str], pos: int) -> Optional[str]:
        """
        Navigate dot delimited tag 'path' into a multi-level dictionary, to return its associated value.

        :param data: Dict, multi-level data dictionary
        :param path: str, dotted JSON tag path
        :param pos: int, zero-based current position in tag path
        :return: string value of the multi-level tag, if available; 'None' otherwise if no tag value found in the path
        """
        tag = path[pos]
        if tag not in data:
            return None

        pos += 1
        if pos == len(path):
            return data[tag]
        else:
            return cls._get_nested_code_value(data[tag], path, pos)

    @classmethod
    def _code_value(cls, code) -> Optional[Tuple[str, str]]:
        """
        Get value of specified dot delimited tag name
        :param code:
        :return: Optional[Tuple[str, str]], 2-tuple of the code type (i.e. info, warning, error) and the
                 validation message template; None if empty code or code unknown in the code dictionary
        """
        if not code:
            return None

        codes: Dict = cls._get_code_dictionary()
        code_path = code.split(".")
        value = cls._get_nested_code_value(codes, code_path, 0)
        if value is not None:
            return code_path[0], value
        else:
            return None

    @staticmethod
    def display(**message):
        assert message and 'code' in message  # should be non-empty, containing a code
        code: str = message.pop('code')
        value: Optional[Tuple[str, str]] = CodeDictionary._code_value(code)
        assert value, f"CodeDictionary.display(): unknown message code {code}"
        message_type, template = value
        code_parts: List[str] = [part.capitalize() for part in code.replace("_", ".").split(".")[1:-1]]
        context: str = ' '.join(code_parts) + ': ' if code_parts else ''
        if message:
            # Message template parameterized with additional named parameter
            # message context, assumed to be referenced by the template
            return f"{message_type.upper()} - {context}{template.format(**message)}"
        else:
            # simple scalar message without parameterization?
            return f"{message_type.upper()} - {context}{template}"


class ReportJsonEncoder(JSONEncoder):
    def default(self, o):
        try:
            iterable = iter(o)
        except TypeError:
            pass
        else:
            return list(iterable)
        # Let the base class default method raise the TypeError
        return JSONEncoder.default(self, o)


def _output(json, flat=False):
    return dumps(json, cls=ReportJsonEncoder, sort_keys=False, indent=None if flat else 4)


class ValidationReporter:
    """
    General wrapper for managing validation status messages: information, warnings and errors.
    The TRAPI version and Biolink Model versions are also tracked for convenience at
    this abstract level although their application is within specific pertinent subclasses.
    """

    # Default major version resolves to latest TRAPI OpenAPI release,
    # specifically 1.3.0, as of September 1st, 2022
    DEFAULT_TRAPI_VERSION = "1"

    _message_type_name: Dict[str, str] = {
        "info": "information",
        "warning": "warnings",
        "error": "errors"
    }

    def __init__(
            self,
            prefix: Optional[str] = None,
            trapi_version: Optional[str] = None,
            biolink_version: Optional[str] = None,
            sources: Optional[Dict] = None,
            strict_validation: bool = False
    ):
        """
        :param prefix: named context of the Validator, used as a prefix in validation messages.
        :type prefix: str
        :param trapi_version: version of component against which to validate the message (mandatory, no default assumed).
        :type trapi_version: str
        :param biolink_version: Biolink Model (SemVer) release against which the knowledge graph is to be
                                validated (Default: if None, use the Biolink Model Toolkit default version.
        :type biolink_version: Optional[str] = None
        :param sources: Dictionary of validation context identifying the ARA and KP for provenance attribute validation
        :type sources: Dict
        :param strict_validation: if True, abstract and mixin elements validate as 'error'; None or False, issue a 'warning'
        :type strict_validation: Optional[bool] = None
        """
        self.prefix: str = prefix + ": " if prefix else ""
        self.trapi_version = trapi_version if trapi_version else latest.get(self.DEFAULT_TRAPI_VERSION)
        self.biolink_version = biolink_version
        self.sources: Optional[Dict] = sources
        self.strict_validation: Optional[bool] = strict_validation
        self.messages: Dict[str, List] = {
            "information": list(),
            "warnings": list(),
            "errors": list()
        }

    def get_trapi_version(self) -> str:
        """
        :return: str, TRAPI (SemVer) version currently targeted by the ValidationReporter.
        """
        return self.trapi_version

    def get_biolink_version(self) -> str:
        """
        :return: Biolink Model version currently targeted by the ValidationReporter.
        :rtype biolink_version: str
        """
        return self.biolink_version

    def is_strict_validation(self) -> bool:
        """
        :return: bool, value of validation strictness set in the ValidationReporter.
        """
        return self.strict_validation

    def has_messages(self) -> bool:
        """Predicate to detect any recorded validation messages.
        :return: bool, True if ValidationReporter has any non-empty messages.
        """
        return self.has_information() or self.has_warnings() or self.has_errors()

    def has_information(self) -> bool:
        """Predicate to detect any recorded information messages.
        :return: bool, True if ValidationReporter has any information messages.
        """
        return bool(self.messages["information"])

    def has_warnings(self) -> bool:
        """Predicate to detect any recorded warning messages.
        :return: bool, True if ValidationReporter has any warning messages.
        """
        return bool(self.messages["warnings"])

    def has_errors(self) -> bool:
        """Predicate to detect any recorded error messages.
        :return: bool, True if ValidationReporter has any error messages.
        """
        return bool(self.messages["errors"])

    def dump_info(self, flat=False) -> str:
        """Dump information messages as JSON.
        :param flat: render output as 'flat' JSON (default: False)
        :return: str, JSON formatted string of information messages.
        """
        return _output(self.messages["information"], flat)

    def dump_warnings(self, flat=False) -> str:
        """Dump warning messages as JSON.
        :param flat: render output as 'flat' JSON (default: False)
        :return: str, JSON formatted string of warning messages.
        """
        return _output(self.messages["warnings"], flat)

    def dump_errors(self, flat=False) -> str:
        """Dump error messages as JSON.
        :param flat: render output as 'flat' JSON (default: False)
        :return: str, JSON formatted string of error messages.
        """
        return _output(self.messages["errors"], flat)

    def dump_messages(self, flat=False) -> str:
        """Dump all messages as JSON.
        :param flat: render output as 'flat' JSON (default: False)
        :return: str, JSON formatted string of all messages.
        """
        return _output(self.messages, flat)

    @staticmethod
    def get_message_type(code: str) -> str:
        """Get type of message code.
        :param code: message code
        :return: message type, one of 'info', 'warning' or 'error'
        """
        code_id_parts: List[str] = code.split('.')
        message_type: str = code_id_parts[0]
        if message_type in ['info', 'warning', 'error']:
            return message_type
        else:
            raise NotImplementedError(
                f"ValidationReport.get_message_type(): {code} is unknown code type: {message_type}"
            )

    def report(self, code: str, **message):
        """
        Capture a single validation message, as per specified 'code' (with any code-specific contextural parameters)
        :param code: 
        :param message: named parameters representing extra (str-formatted) context for the given code message
        :return: None (internally record the validation message)
        """
        message_type = self.get_message_type(code)
        message_set = self._message_type_name[message_type]
        message['code'] = code  # add the code into the message
        self.messages[message_set].append(message)

    def add_messages(self, new_messages: Dict[str, List]):
        """
        Batch addition of a dictionary of messages to a ValidationReporter instance.

        :param new_messages: Dict[str, List], with key one of
                             "information", "warnings" or "errors",
                              with Lists of (structured) messages.
        """
        for key in self.messages:
            if key in new_messages:
                self.messages[key].extend(new_messages[key])

    def get_messages(self) -> Dict[str, List[Dict]]:
        """
        Get copy of all messages as a Python data structure.
        :return: Dict (copy) of all validation messages in the ValidationReporter.
        """
        return copy.deepcopy(self.messages)

    def get_info(self) -> List:
        """
        Get copy of all recorded information messages.
        :return: List, copy of all information messages.
        """
        return copy.deepcopy(self.messages["information"])

    def get_warnings(self) -> List:
        """
        Get copy of all recorded warning messages.
        :return: List, copy of all warning messages.
        """
        return copy.deepcopy(self.messages["warnings"])

    def get_errors(self) -> List:
        """
        Get copy of all recorded error messages.
        :return: List, copy of all error messages.
        """
        return copy.deepcopy(self.messages["errors"])

    ############################
    # General Instance methods #
    ############################
    def merge(self, reporter):
        """
        Merge all messages and metadata from a second ValidationReporter,
        into the calling ValidationReporter instance.

        :param reporter: second ValidationReporter
        """
        assert isinstance(reporter, ValidationReporter)

        # new coded messages also need to be merged!
        self.add_messages(reporter.get_messages())

        # First come, first serve... We only overwrite
        # empty versions in the parent reporter
        if not self.trapi_version:
            self.trapi_version = reporter.trapi_version
        if not self.biolink_version:
            self.biolink_version = reporter.biolink_version

    def to_dict(self) -> Dict:
        """
        Export ValidationReporter contents as a Python dictionary
        (including TRAPI version, Biolink Model version and messages)
        :return: Dict
        """
        return {
            "trapi_version": self.trapi_version,
            "biolink_version": self.biolink_version,
            "messages": self.get_messages()
        }

    def apply_validation(self, validation_method, *args, **kwargs) -> bool:
        """
        Wrapper to allow validation_methods direct access to the ValidationReporter.

        :param validation_method: function which accepts this instance of the
               ValidationReporter as its first argument, for use in reporting validation errors.
        :param args: any positional arguments to the validation_method, after the initial ValidationReporter argument
        :param kwargs: any (optional, additional) keyword arguments to the validation_method, after positional arguments
        :return: bool, returns 'False' if validation method documented errors; True otherwise
        """
        validation_method(self, *args, **kwargs)
        if self.has_errors():
            return False
        else:
            return True

    @staticmethod
    def has_validation_errors(tag: str, case: Dict) -> bool:
        """
        Check if test case has validation errors.
        :param tag: str, top level string key in the 'case' whose value is the validation messages 'dictionary'
        :param case: Dict, containing error messages in a structurally similar
                     format to what is returned by the to_dict() method in this class.
        :return: True if the case contains validation messages
        """
        #
        # The 'case' dictionary object could have a format something like this:
        #
        #     tag: {
        #         "trapi_version": "1.3",
        #         "biolink_version": "3.0.2",
        #         "messages": {
        #             "information": [],
        #             "warnings": [
        #                 {
        #                     'code': "warning.predicate.non_canonical",
        #                     'predicate': "biolink:participates_in"
        #                 }
        #             ],
        #             "errors": [
        #                 {
        #                     'code': "error.knowledge_graph.empty_nodes"
        #                 }
        #             ]
        #         }
        #     }
        #
        # where 'tag' == 'messages' and we have a non-empty "errors" set of messages
        #
        if case is not None and tag in case and \
                'messages' in case[tag] and \
                'errors' in case[tag]['messages'] and \
                case[tag]['messages']['errors']:
            return True
        else:
            return False

    def display(self, **message) -> str:
        """
        Augmented message display wrapper prepends
        the ValidationReporter prefix to a
        resolved coded validation message.
        :return: str, fully rendered validation message
        """
        return self.prefix + CodeDictionary.display(**message)



"""
Version-specific Biolink Model semantic validation of knowledge graph components.
"""
from typing import Optional, Any, Dict, List, Tuple, Set
from enum import Enum
from functools import lru_cache
from urllib.error import HTTPError
from pprint import PrettyPrinter
import logging

from bmt import Toolkit
from linkml_runtime.linkml_model import ClassDefinition, Element

from reasoner_validator.sri.util import is_curie
from reasoner_validator.report import ValidationReporter
from reasoner_validator.versioning import SemVer, SemVerError

logger = logging.getLogger(__name__)

pp = PrettyPrinter(indent=4)

# TODO: is there a better way to ensure that a Biolink Model compliance test
#       runs quickly enough if the knowledge graph is very large?
#       Limiting nodes and edges viewed may miss deeply embedded errors(?)
_MAX_TEST_NODES = 1000
_MAX_TEST_EDGES = 100


def _get_biolink_model_schema(biolink_version: Optional[str] = None) -> Optional[str]:
    # Get Biolink Model Schema
    if biolink_version:
        try:
            svm = SemVer.from_string(biolink_version, ignore_prefix='v')

            # Sanity check: override SemVer object to ignore prerelease and
            # buildmetadata variants of the Biolink Version given
            svm = SemVer(major=svm.major, minor=svm.minor, patch=svm.patch)
        except SemVerError:
            raise TypeError(
                "The 'biolink_version' argument '"
                + biolink_version
                + "' is not a properly formatted semantic version?"
            )

        if svm >= SemVer.from_string("2.2.14"):
            biolink_version = "v" + str(svm)
        else:
            biolink_version = str(svm)
        schema = f"https://raw.githubusercontent.com/biolink/biolink-model/{biolink_version}/biolink-model.yaml"
        return schema
    else:
        return None


# At any given time, only a modest number of Biolink Model versions
# are expected to be active targets for SRI Test validations?
@lru_cache(maxsize=3)
def get_biolink_model_toolkit(biolink_version: Optional[str] = None) -> Toolkit:
    """
    Return Biolink Model Toolkit corresponding to specified version of the model (Default: current 'latest' version).

    :param biolink_version: Optional[str], caller specified Biolink Model version (default: None)
    :type biolink_version: Optional[str] or None
    :return: Biolink Model Toolkit.
    :rtype: Toolkit

    """
    if biolink_version:
        # If errors occur while instantiating non-default Toolkit;
        # then log the error but just use default as a workaround?
        try:
            biolink_schema = _get_biolink_model_schema(biolink_version=biolink_version)
            bmt = Toolkit(biolink_schema)
            return bmt
        except (TypeError, HTTPError) as ex:
            logger.error(str(ex))

    # 'latest' default Biolink Model
    # version of given Toolkit returned
    return Toolkit()


class TRAPIGraphType(Enum):
    """ Enum type of Biolink Model compliant graph data being validated."""
    Input_Edge = "Input Edge"
    Query_Graph = "Query Graph"
    Knowledge_Graph = "Knowledge Graph"


class BiolinkValidator(ValidationReporter):
    """
    Wrapper class for Biolink Model validation.
    """
    def __init__(
        self,
        graph_type: TRAPIGraphType,
        biolink_version: Optional[str] = None,
        sources: Optional[Dict[str, str]] = None,
        strict_validation: bool = False
    ):
        """
        Biolink Validator constructor.

        :param graph_type: type of graph data being validated
        :type graph_type: TRAPIGraphType
        :param biolink_version: caller specified Biolink Model version (default: None)
        :type biolink_version: Optional[str] or None
        :param sources: Dictionary of validation context identifying the ARA and KP for provenance attribute validation
        :type sources: Optional[Dict[str,str]]
        """
        self.bmt: Toolkit = get_biolink_model_toolkit(biolink_version=biolink_version)
        resolved_biolink_version = self.bmt.get_model_version()
        ValidationReporter.__init__(
            self,
            prefix=f"Biolink Validation of {graph_type.value}",
            biolink_version=resolved_biolink_version,
            sources=sources,
            strict_validation=strict_validation
        )
        self.graph_type: TRAPIGraphType = graph_type
        self.nodes: Set[str] = set()

    def minimum_required_biolink_version(self, version: str) -> bool:
        """
        :param version: simple 'major.minor.patch' Biolink Model SemVer
        :return: True if current version is equal to, or newer than, a targeted 'minimum_version'
        """
        try:
            current: SemVer = SemVer.from_string(self.biolink_version)
            target: SemVer = SemVer.from_string(version)
            return current >= target
        except SemVerError as sve:
            logger.error(f"minimum_required_biolink_version() error: {str(sve)}")
            return False

    def get_result(self) -> Tuple[str, Optional[Dict[str, List[Dict]]]]:
        """
        Get result of validation.

        :return: model version of the validation and dictionary of reported validation messages.
        :rtype Tuple[str, Optional[Dict[str, Set[str]]]]
        """
        return self.bmt.get_model_version(), self.get_messages()

    def validate_graph_node(self, node_id: str, slots: Dict[str, Any]):
        """
        Validate slot properties (mainly 'categories') of a node.

        :param node_id: identifier of a concept node
        :type node_id: str, node identifier
        :param slots: properties of the node
        :type slots: Dict
        """
        logger.debug(f"{node_id}: {str(slots)}")

        if self.graph_type is TRAPIGraphType.Knowledge_Graph:
            # TODO: this will fail for an earlier TRAPI data schema version
            #       which didn't use the tag 'categories' for nodes...
            #       probably no longer relevant to the community?
            if 'categories' in slots:
                if not isinstance(slots["categories"], List):
                    self.report(code="error.knowledge_graph.node.categories.not_array", node_id=node_id)
                else:
                    categories = slots["categories"]
                    node_prefix_mapped: bool = False
                    for category in categories:
                        category: Optional[ClassDefinition] = \
                            self.validate_category(
                                context="knowledge_graph",
                                node_id=node_id,
                                category=category
                            )
                        if category:
                            possible_subject_categories = self.bmt.get_element_by_prefix(node_id)
                            if category.name in possible_subject_categories:
                                node_prefix_mapped = True
                    if not node_prefix_mapped:
                        self.report(
                            code="warning.knowledge_graph.node.unmapped_prefix",
                            node_id=node_id, categories=str(categories)
                        )
            else:
                self.report(
                    code="error.knowledge_graph.node.category.missing",
                    context=self.graph_type.value, node_id=node_id
                )

            # TODO: Do we need to (or can we) validate here, any other Knowledge Graph node fields? Perhaps not yet?

        else:  # Query Graph node validation

            if "ids" in slots and slots["ids"]:
                ids = slots["ids"]
                if not isinstance(ids, List):
                    self.report(code="error.query_graph.node.ids.not_array", node_id=node_id)
                    # we'll pretend that the ids were mistakenly
                    # just a scalar string, then continue validating
                    ids = [ids]
            else:
                ids = list()  # a null "ids" value is permitted in QNodes

            if "categories" in slots:
                categories = slots["categories"]
                if categories:
                    if not isinstance(categories, List):
                        self.report(code="error.query_graph.node.categories.not_array", node_id=node_id)
                    else:
                        id_prefix_mapped: Dict = {identifier: False for identifier in ids}
                        for category in categories:
                            # category validation may report an error internally
                            category: Optional[ClassDefinition] = \
                                self.validate_category(
                                    context="query_graph",
                                    node_id=node_id,
                                    category=category
                                )
                            if category:
                                for identifier in ids:  # may be empty list if not provided...
                                    possible_subject_categories = self.bmt.get_element_by_prefix(identifier)
                                    if category.name in possible_subject_categories:
                                        id_prefix_mapped[identifier] = True
                        unmapped_ids = [
                            identifier for identifier in id_prefix_mapped.keys() if not id_prefix_mapped[identifier]
                        ]
                        if unmapped_ids:
                            self.report(
                                code="warning.query_graph.node.ids.unmapped_to_categories",
                                node_id=node_id,
                                unmapped_ids=str(unmapped_ids),
                                categories=str(categories)
                            )

                # else:  # null "categories" value is permitted in QNodes by nullable: true
            # else:  # missing "categories" key is permitted in QNodes by nullable: true

            if 'is_set' in slots:
                is_set = slots["is_set"]
                if is_set and not isinstance(is_set, bool):
                    self.report(code="error.query_graph.node.is_set.not_boolean", node_id=node_id)
            # else:  # a missing key or null "is_set" value is permitted in QNodes but defaults to 'False'

            # constraints  # TODO: how do we validate node constraints?
            pass

    def set_nodes(self, nodes: Set):
        self.nodes.update(nodes)

    def validate_element_status(self, context: str, name: str) -> Optional[Element]:
        """
        Detect element missing from Biolink, or is deprecated, abstract or mixin, signalled as a failure or warning.

        :param context: parsing context (e.g. 'Node')
        :param name: name of putative Biolink element ('class')

        :return: Optional[Element], Biolink Element resolved to 'name' if element passed all validation; None otherwise.
        """
        element: Optional[Element] = self.bmt.get_element(name)
        if not element:
            self.report(code=f"error.{context}.unknown", name=name)
        elif element.deprecated:
            self.report(code=f"warning.{context}.deprecated", name=name)
            return None
        elif element.abstract:
            if self.strict_validation:
                self.report(code=f"error.{context}.abstract",  name=name)
            else:
                self.report(code=f"info.{context}.abstract", name=name)
            return None
        elif self.bmt.is_mixin(name):
            # A mixin cannot be instantiated thus it should not be given as an input concept category
            if self.strict_validation:
                self.report(code=f"error.{context}.mixin", name=name)
            else:
                self.report(code=f"info.{context}.mixin", name=name)
            return None
        else:
            return element

    def validate_attributes(self, edge: Dict):
        """
        :param edge: Dict, the edge object associated with some attributes are expected to be found
        :return: None (validation messages captured in the 'self' BiolinkValidator context)
        """
        if 'attributes' not in edge.keys():
            self.report(code="error.knowledge_graph.edge.attribute.missing")
        elif not edge['attributes']:
            self.report(code="error.knowledge_graph.edge.attribute.empty")
        elif not isinstance(edge['attributes'], List):
            self.report(code="error.knowledge_graph.edge.attribute.not_list")
        else:
            attributes = edge['attributes']

            ara_source: Optional[str] = None
            kp_source: Optional[str] = None
            kp_source_type: Optional[str] = None
            if self.sources:
                if 'ara_source' in self.sources and self.sources['ara_source']:
                    ara_source: str = self.sources['ara_source']
                    if not ara_source.startswith("infores:"):
                        ara_source = f"infores:{ara_source}"
                if 'kp_source' in self.sources and self.sources['kp_source']:
                    kp_source: str = self.sources['kp_source']
                    if not kp_source.startswith("infores:"):
                        kp_source = f"infores:{kp_source}"
                kp_source_type = self.sources['kp_source_type'] \
                    if 'kp_source_type' in self.sources and self.sources['kp_source_type'] else 'aggregator'
                kp_source_type = f"biolink:{kp_source_type}_knowledge_source"

            # Expecting ARA and KP 'aggregator_knowledge_source' attributes?
            found_ara_knowledge_source = False
            found_kp_knowledge_source = False
            found_primary_or_original_knowledge_source = False

            for attribute in attributes:

                # Validate attribute_type_id
                if 'attribute_type_id' not in attribute:
                    self.report(code="error.knowledge_graph.edge.attribute.type_id.missing")
                elif not attribute['attribute_type_id']:
                    self.report(code="error.knowledge_graph.edge.attribute.type_id.empty")
                elif 'value' not in attribute:
                    self.report(code="error.knowledge_graph.edge.attribute.value.missing")
                elif not attribute['value']:
                    self.report(code="error.knowledge_graph.edge.attribute.value.empty")
                else:
                    attribute_type_id: str = attribute['attribute_type_id']
                    value = attribute['value']

                    # TODO: there seems to be non-uniformity in provenance attribute values for some KP/ARA's
                    #       in which a value is returned as a Python list (of at least one element?) instead
                    #       of a string. Here, to ensure full coverage of the attribute values returned,
                    #       we'll coerce scalar values into a list, then iterate.
                    if not isinstance(value, List):
                        value = [value]

                    if not is_curie(attribute_type_id):
                        self.report(
                            code="error.knowledge_graph.edge.attribute.type_id.not_curie",
                            attribute_type_id=str(attribute_type_id)
                        )
                    else:
                        # 'attribute_type_id' is a CURIE, but how well does it map?
                        prefix = attribute_type_id.split(":", 1)[0]
                        if prefix == 'biolink':
                            biolink_class = self.validate_element_status(
                                context="knowledge_graph.attribute.type_id",
                                name=attribute_type_id
                            )
                            if biolink_class:
                                if not self.bmt.is_association_slot(attribute_type_id):
                                    self.report(
                                        code="warning.knowledge_graph.edge.attribute.type_id.not_association_slot",
                                        attribute_type_id=str(attribute_type_id)
                                    )

                                else:
                                    # it is a Biolink 'association_slot' but now, validate what kind?

                                    # TODO: only check knowledge_source provenance here for now.
                                    #       Are there other association_slots to be validated here too?

                                    # Check Edge Provenance attributes
                                    if attribute_type_id not in \
                                            [
                                                "biolink:aggregator_knowledge_source",
                                                "biolink:primary_knowledge_source",
                                                "biolink:original_knowledge_source"

                                            ]:

                                        # TODO: not interested in any other attribute_type_id's at this moment
                                        continue

                                    if attribute_type_id in \
                                            [
                                                "biolink:primary_knowledge_source",
                                                "biolink:original_knowledge_source"
                                            ]:
                                        found_primary_or_original_knowledge_source = True

                                    # ... now, check the infores values against various expectations
                                    for infores in value:
                                        if not infores.startswith("infores:"):
                                            self.report(
                                                code="error.knowledge_graph.edge.provenance.infores.missing",
                                                infores=infores
                                            )
                                        else:
                                            if ara_source and \
                                                    attribute_type_id == "biolink:aggregator_knowledge_source" and \
                                                    infores == ara_source:
                                                found_ara_knowledge_source = True
                                            elif kp_source and \
                                                    attribute_type_id == kp_source_type and \
                                                    infores == kp_source:
                                                found_kp_knowledge_source = True

                        # if not a Biolink association_slot, at least,
                        # check if it is an id prefix known to Biolink.
                        # We won't call it a hard error, but issue a warning
                        elif not self.bmt.get_element_by_prefix(prefix):
                            self.report(
                                code="warning.knowledge_graph.edge.attribute.type_id.unknown_prefix",
                                attribute_type_id=str(attribute_type_id)
                            )

                        else:
                            self.report(
                                code="info.attribute.type_id.non_biolink_prefix",
                                attribute_type_id=str(attribute_type_id)
                            )

            # TODO: After all the attributes have been scanned, check for provenance. Treat as warnings for now
            if ara_source and not found_ara_knowledge_source:
                self.report(code="warning.knowledge_graph.edge.provenance.ara.missing")

            if kp_source and not found_kp_knowledge_source:
                self.report(
                    code="warning.knowledge_graph.edge.provenance.kp.missing",
                    kp_source=kp_source,
                    kp_source_type=kp_source_type
                )

            if not found_primary_or_original_knowledge_source:
                self.report(code="warning.knowledge_graph.edge.provenance.missing_primary")

    def validate_predicate(self, edge_id: str, predicate: str):
        """
        :param predicate: putative Biolink Model predicate to be validated
        :return:
        """
        context: str = f"{self.graph_type.name.lower()}.predicate"
        # Validate the putative predicate as *not* being abstract, deprecated or a mixin
        biolink_class = self.validate_element_status(
            context=context,
            name=predicate
        )
        if biolink_class:
            if not self.bmt.is_predicate(predicate):
                self.report(
                    code=f"error.{context}.invalid",
                    context=self.graph_type.value,
                    edge_id=edge_id,
                    predicate=predicate
                )
            elif self.minimum_required_biolink_version("2.2.0") and \
                    not self.bmt.is_translator_canonical_predicate(predicate):
                self.report(
                    code=f"warning.{context}.non_canonical",
                    context=self.graph_type.value,
                    edge_id=edge_id,
                    predicate=predicate
                )

    def validate_graph_edge(self, edge: Dict):
        """
        Validate slot properties of a relationship ('biolink:Association') edge.

        :param edge: dictionary of slot properties of the edge.
        :type edge: dict[str, str]
        """
        # logger.debug(edge)
        # edge data fields to be validated...
        subject_id = edge['subject'] if 'subject' in edge else None

        predicates = predicate = None
        if self.graph_type is TRAPIGraphType.Knowledge_Graph:
            predicate = edge['predicate'] if 'predicate' in edge else None
            edge_label = predicate
        else:
            # Query Graph...
            predicates = edge['predicates'] if 'predicates' in edge else None
            edge_label = str(predicates)

        object_id = edge['object'] if 'object' in edge else None

        edge_id = f"{str(subject_id)}--{edge_label}->{str(object_id)}"

        context: str = self.graph_type.name.lower()

        if not subject_id:
            self.report(code=f"error.{context}.edge.subject.missing", edge_id=edge_id)
        elif subject_id not in self.nodes:
            self.report(code=f"error.{context}.edge.subject.missing_from_nodes", object_id=subject_id)
        if self.graph_type is TRAPIGraphType.Knowledge_Graph:
            if not predicate:
                self.report(
                    code="error.knowledge_graph.predicate.missing",
                    context=self.graph_type.value, edge_id=edge_id
                )
            else:
                self.validate_predicate(edge_id=edge_id, predicate=predicate)
        else:  # is a Query Graph...
            if predicates is None:
                # Query Graphs can have a missing or null predicates slot
                pass
            elif not isinstance(predicates, List):
                self.report(code="error.query_graph.predicate.not_array", edge_id=edge_id)
            elif len(predicates) == 0:
                self.report(code="error.query_graph.predicate.empty_array", edge_id=edge_id)
            else:
                # Should now be a non-empty list of CURIES which are valid Biolink Predicates
                for predicate in predicates:
                    if not predicate:
                        continue  # sanity check
                    self.validate_predicate(edge_id=edge_id, predicate=predicate)
        if not object_id:
            self.report(code=f"error.{context}.edge.object.missing", edge_id=edge_id)
        elif object_id not in self.nodes:
            self.report(code=f"error.{context}.edge.object.missing_from_nodes", object_id=object_id)
        if self.graph_type is TRAPIGraphType.Knowledge_Graph:
            self.validate_attributes(edge=edge)
        else:
            # TODO: do we need to validate Query Graph 'constraints' slot contents here?
            pass

    def validate_category(
            self,
            context: str,
            node_id: Optional[str],
            category: Optional[str]
    ) -> ClassDefinition:
        """
        Validate a Biolink category.

        :param context: str, label for context of concept whose category is being validated, i.e. 'Subject' or 'Object'
        :param node_id: str, CURIE of concept node whose category is being validated
        :param category: str, CURIE of putative concept 'category'
        :return:
        """
        biolink_class: Optional[ClassDefinition] = None
        if category:
            # this method reports unknown Biolink Elements
            biolink_class = self.validate_element_status(
                    context=f"{context}.node.category",
                    name=category
            )
            if biolink_class and not self.bmt.is_category(category):
                self.report(code=f"error.{context}.node.category.unknown", category=category)
                biolink_class = None
        else:
            self.report(code=f"error.{context}.node.category.missing", node_id=node_id)

        return biolink_class

    def validate_input_edge_node(self, context: str, node_id: Optional[str], category: Optional[str]):
        if node_id:
            biolink_class: Optional[ClassDefinition] = self.validate_category(
                context="input_edge",
                node_id=node_id,
                category=category
            )
            if biolink_class:
                possible_subject_categories = self.bmt.get_element_by_prefix(node_id)
                if biolink_class.name not in possible_subject_categories:
                    self.report(
                        code="warning.input_edge.node.id.unmapped_to_category",
                        context=context,
                        node_id=node_id,
                        category=category
                    )
            # else, we will have already reported an error in validate_category()
        else:
            self.report(code="error.input_edge.node.id.missing", context=context)

    def check_biolink_model_compliance_of_input_edge(self, edge: Dict[str, str]):
        """
        Validate a templated test input edge contents against the current BMT Biolink Model release.

        Sample method 'edge' with expected dictionary tags:

        {
            'subject_category': 'biolink:AnatomicalEntity',
            'object_category': 'biolink:AnatomicalEntity',
            'predicate': 'biolink:subclass_of',
            'subject': 'UBERON:0005453',
            'object': 'UBERON:0035769'
        }

        :param edge: basic dictionary of a templated input edge - S-P-O including concept Biolink Model categories
        :type edge: Dict[str,str]
        """
        # data fields to be validated...
        subject_category_curie = edge['subject_category'] if 'subject_category' in edge else None
        object_category_curie = edge['object_category'] if 'object_category' in edge else None
        predicate = edge['predicate'] if 'predicate' in edge else None
        subject_curie = edge['subject'] if 'subject' in edge else None
        object_curie = edge['object'] if 'object' in edge else None

        edge_id = f"{str(subject_curie)}--{predicate}->{str(object_curie)}"

        self.validate_input_edge_node(
            context='Subject',
            node_id=subject_curie,
            category=subject_category_curie
        )
        if not predicate:
            self.report(
                code="error.input_edge.predicate.missing",
                edge_id=edge_id
            )
        else:
            self.validate_predicate(edge_id=edge_id, predicate=predicate)

        self.validate_input_edge_node(
            context='Object',
            node_id=object_curie,
            category=object_category_curie
        )

    def check_biolink_model_compliance(self, graph: Dict):
        """
        Validate a TRAPI-schema compliant Message graph-like data structure
        against the currently active Biolink Model Toolkit model version.
    
        :param graph: knowledge graph to be validated
        :type graph: Dict
        """
        if not graph:
            self.report(code="warning.graph.empty", context=self.graph_type.value)
            return  # nothing really more to do here!

        # Access graph data fields to be validated
        nodes: Optional[Dict]
        if 'nodes' in graph and graph['nodes']:
            nodes = graph['nodes']
        else:
            # Query Graphs can have an empty nodes catalog
            if self.graph_type is not TRAPIGraphType.Query_Graph:
                self.report(code="error.knowledge_graph.nodes.empty")
            # else:  Query Graphs can omit the 'nodes' tag
            nodes = None

        edges: Optional[Dict]
        if 'edges' in graph and graph['edges']:
            edges = graph['edges']
        else:
            if self.graph_type is not TRAPIGraphType.Query_Graph:
                self.report(code="error.knowledge_graph.edges.empty")
            # else:  Query Graphs can omit the 'edges' tag
            edges = None

        # I only do a sampling of node and edge content. This ensures that
        # the tests are performant but may miss errors deeper inside the graph?
        nodes_seen = 0
        if nodes:
            for node_id, details in nodes.items():

                self.validate_graph_node(node_id, details)

                nodes_seen += 1
                if nodes_seen >= _MAX_TEST_NODES:
                    break

            # Needed for the subsequent edge validation
            self.set_nodes(set(nodes.keys()))

            edges_seen = 0
            if edges:
                for edge in edges.values():

                    # print(f"{str(edge)}", flush=True)
                    self.validate_graph_edge(edge)

                    edges_seen += 1
                    if edges_seen >= _MAX_TEST_EDGES:
                        break


def check_biolink_model_compliance_of_input_edge(
        edge: Dict[str, str],
        biolink_version: Optional[str] = None,
        strict_validation: Optional[bool] = None
) -> BiolinkValidator:
    """
    Validate an (SRI Testing style) input edge record
    against a designated Biolink Model release.

    Sample method 'edge' with expected dictionary tags:

    {
        'subject_category': 'biolink:AnatomicalEntity',
        'object_category': 'biolink:AnatomicalEntity',
        'predicate': 'biolink:subclass_of',
        'subject': 'UBERON:0005453',
        'object': 'UBERON:0035769'
    }

    :param edge: basic contents of a templated input edge - S-P-O including concept Biolink Model categories
    :type edge: Dict[str,str]
    :param biolink_version: Biolink Model (SemVer) release against which the knowledge graph is to be
                            validated (Default: if None, use the Biolink Model Toolkit default version.
    :type biolink_version: Optional[str] = None
    :param strict_validation: if True, abstract and mixin elements validate as 'error'; False, issue 'info' message.
    :type strict_validation: bool = False

    :returns: Biolink Model validator cataloging validation messages (may be empty)
    :rtype: BiolinkValidator
    """
    if strict_validation is None:
        # Test Input Edges are like Query Graph data,
        # hence, abstract and mixins ARE permitted
        strict_validation = False
    validator = BiolinkValidator(
        graph_type=TRAPIGraphType.Input_Edge,
        biolink_version=biolink_version,
        strict_validation=strict_validation
    )
    validator.check_biolink_model_compliance_of_input_edge(edge)
    return validator


def check_biolink_model_compliance_of_query_graph(
        graph: Dict,
        biolink_version: Optional[str] = None,
        strict_validation: Optional[bool] = None
) -> BiolinkValidator:
    """
    Validate a TRAPI-schema compliant Message Query Graph
    against a designated Biolink Model release.

    Since a Query graph is usually an incomplete knowledge graph specification,
    the validation undertaken is not 'strict'

    :param graph: query graph to be validated
    :type graph: Dict
    :param biolink_version: Biolink Model (SemVer) release against which the knowledge graph is to be
                            validated (Default: if None, use the Biolink Model Toolkit default version.
    :type biolink_version: Optional[str] = None
    :param strict_validation: if True, abstract and mixin elements validate as 'error'; False, issue 'info' message.
    :type strict_validation: Optional[bool] = None; defaults to 'False' if not set

    :returns: Biolink Model validator cataloging validation messages (may be empty)
    :rtype: BiolinkValidator
    """
    # One typically won't be stringent in QueryGraph validation; however,
    # the strict_validation flag is set to a default of 'False' only if it is NOT set
    if strict_validation is None:
        # Query Graph data can use abstract and mixins
        strict_validation = False
    validator = BiolinkValidator(
        graph_type=TRAPIGraphType.Query_Graph,
        biolink_version=biolink_version,
        strict_validation=strict_validation
    )
    validator.check_biolink_model_compliance(graph)
    return validator


def check_biolink_model_compliance_of_knowledge_graph(
    graph: Dict,
    biolink_version: Optional[str] = None,
    sources: Optional[Dict] = None,
    strict_validation: Optional[bool] = None
) -> BiolinkValidator:
    """
    Strict validation of a TRAPI-schema compliant Message Knowledge Graph
     against a designated Biolink Model release.

    :param graph: knowledge graph to be validated.
    :type graph: Dict
    :param biolink_version: Biolink Model (SemVer) release against which the knowledge graph is to be
                            validated (Default: if None, use the Biolink Model Toolkit default version.
    :type biolink_version: Optional[str] = None
    :param sources: Dictionary of validation context identifying the ARA and KP for provenance attribute validation
    :type sources: Dict
    :param strict_validation: if True, abstract and mixin elements validate as 'error'; False, issue 'info' message.
    :type strict_validation: Optional[bool] = None; defaults to 'True' if not set

    :returns: Biolink Model validator cataloging validation messages (may be empty)
    :rtype: BiolinkValidator
    """
    # One typically will want stringent validation for Knowledge Graphs; however,
    # the strict_validation flag is set to a default of 'True' only if it is NOT set
    if strict_validation is None:
        # Knowledge Graphs generally ought NOT to use abstract and mixins
        strict_validation = True
    validator = BiolinkValidator(
        graph_type=TRAPIGraphType.Knowledge_Graph,
        biolink_version=biolink_version,
        sources=sources,
        strict_validation=strict_validation
    )
    validator.check_biolink_model_compliance(graph)
    return validator



"""TRAPI Validation Functions."""
from typing import Optional
import copy
from functools import lru_cache

import jsonschema
import requests

from yaml import load, CLoader as Loader

from reasoner_validator.report import ValidationReporter
from reasoner_validator.trapi.mapping import check_node_edge_mappings
from reasoner_validator.versioning import latest, versions, GIT_ORG, GIT_REPO


@lru_cache()
def _load_schema(trapi_version: str):
    """Load schema from GitHub."""
    result = requests.get(
        f"https://raw.githubusercontent.com/{GIT_ORG}/{GIT_REPO}/v{trapi_version}/TranslatorReasonerAPI.yaml"
    )
    spec = load(result.text, Loader=Loader)
    components = spec["components"]["schemas"]
    for component, schema in components.items():
        openapi_to_jsonschema(schema)
    schemas = dict()
    for component in components:
        # build json schema against which we validate
        subcomponents = copy.deepcopy(components)
        schema = subcomponents.pop(component)
        schema["components"] = {"schemas": subcomponents}
        schemas[component] = schema
    return schemas


def load_schema(trapi_version: str):
    """Load schema from GitHub."""
    full_version = latest.get(trapi_version)
    if full_version not in versions:
        raise ValueError(f"No TRAPI version {trapi_version}")
    return _load_schema(full_version)


def fix_nullable(schema) -> None:
    """Fix nullable schema."""
    if "oneOf" in schema:
        schema["oneOf"].append({"type": "null"})
        return
    if "anyOf" in schema:
        schema["anyOf"].append({"type": "null"})
        return
    schema["oneOf"] = [
        {
            key: schema.pop(key)
            for key in list(schema.keys())
        },
        {"type": "null"},
    ]


def openapi_to_jsonschema(schema) -> None:
    """Convert OpenAPI schema to JSON schema."""
    if "allOf" in schema:
        # September 1, 2022 hacky patch to rewrite 'allOf' tagged subschemata to 'oneOf'
        # TODO: TRAPI needs to change this in release 1.4
        schema["oneOf"] = schema.pop("allOf")
    if schema.get("type", None) == "object":
        for tag, prop in schema.get("properties", dict()).items():
            openapi_to_jsonschema(prop)
    if schema.get("type", None) == "array":
        openapi_to_jsonschema(schema.get("items", dict()))
    if schema.pop("nullable", False):
        fix_nullable(schema)


class TRAPISchemaValidator(ValidationReporter):
    """
    TRAPI Validator is a wrapper class for validating
    conformance of JSON messages to the Translator Reasoner API.
    """
    def __init__(
            self,
            trapi_version: Optional[str] = None
    ):
        """
        TRAPI Validator constructor.

        Parameters
        ----------
        trapi_version : str
            version of component to validate against
        """
        resolved_biolink_version = latest.get(trapi_version)
        ValidationReporter.__init__(
            self,
            prefix="TRAPI Validation",
            trapi_version=resolved_biolink_version
        )

    def validate(self, instance, component):
        """Validate instance against schema.

        Parameters
        ----------
        instance
            dict, instance to validate
        component : str
            str, TRAPI subschema to validate (e.g. 'Query', 'QueryGraph', 'KnowledgeGraph', 'Result'; Default: 'Query')

        Raises
        ------
        `ValidationError <https://python-jsonschema.readthedocs.io/en/latest/errors/#jsonschema.exceptions.ValidationError>`_
            If the instance is invalid.

        Examples
        --------
        >>> TRAPISchemaValidator(trapi_version="1.3.0").validate({"message": {}}, "QGraph")

        """
        schema = load_schema(self.trapi_version)[component]
        jsonschema.validate(instance, schema)

    def is_valid_trapi_query(self, instance, component: str = "Query"):
        """Make sure that the Message is a syntactically valid TRAPI Query JSON object.

        Parameters
        ----------
        instance:
            Dict, instance to validate
        component:
            str, TRAPI subschema to validate (e.g. 'Query', 'QueryGraph', 'KnowledgeGraph', 'Result'; Default: 'Query')

        Returns
        -------
        Validation ("information", "warning" and "error") messages are returned within the host TRAPIValidator instance.

        Examples
        --------
        >>> TRAPISchemaValidator(trapi_version="1.3.0").is_valid_trapi_query({"message": {}}, component="Query")
        """
        try:
            self.validate(
                instance=instance,
                component=component
            )
        except jsonschema.ValidationError as e:
            self.report(code="error.trapi.validation", trapi_version=self.trapi_version, exception=e.message)


def check_trapi_validity(instance, trapi_version: str, component: str = "Query") -> TRAPISchemaValidator:
    """
    Checks schema compliance of a Query component against a given TRAPI version.

    Parameters
    ----------
    instance:
        Dict, of format {"message": {}}
    component:
        str, TRAPI subschema to validate (e.g. 'Query', 'QueryGraph', 'KnowledgeGraph', 'Result'; Default: 'Query')
    trapi_version:
        str, version of component to validate against

    Returns
    -------
    ValidationReporter catalog of "information", "warnings" or "errors" indexed messages (may be empty)
    """
    trapi_validator = TRAPISchemaValidator(trapi_version=trapi_version)
    trapi_validator.is_valid_trapi_query(instance, component=component)
    return trapi_validator



from typing import Dict

from reasoner_validator import ValidationReporter


class MappingValidator(ValidationReporter):
    """
    The Mapping Validator is a wrapper class for detecting
    dangling references between nodes and edges of a graph.
    This is more of a TRAPI expectation (that all nodes and edges identifiers refer to one another)
    """
    def __init__(self):
        """
        Mapping Validator constructor.
        """
        ValidationReporter.__init__(
            self,
            prefix="Validating Knowledge Graph Node and Edge Mappings"
        )

    def check_dangling_references(self, graph: Dict):
        if not ('nodes' in graph and graph['nodes'] and 'edges' in graph and graph['edges']):
            self.report(code="warning.graph.empty")
        else:
            pass


# Detect 'dangling nodes/edges' by iterating through node <-> edge mappings)
def check_node_edge_mappings(graph: Dict) -> MappingValidator:
    validator: MappingValidator = MappingValidator()
    validator.check_dangling_references(graph)
    return validator



from typing import Optional, List, Dict

from reasoner_validator.report import ValidationReporter
from reasoner_validator.biolink import (
    check_biolink_model_compliance_of_query_graph,
    check_biolink_model_compliance_of_knowledge_graph,
    BiolinkValidator
)

# Maximum number of data points to scrutinize
# in various parts TRAPI Query Response.Message
from reasoner_validator.trapi import check_trapi_validity, TRAPISchemaValidator
from reasoner_validator.trapi.mapping import MappingValidator, check_node_edge_mappings

TEST_DATA_SAMPLE_SIZE = 10


class TRAPIResponseValidator(ValidationReporter):
    """
    TRAPI Validator is an overall wrapper class for validating
    conformance of TRAPI Responses to TRAPI and the Biolink Model.
    """

    def __init__(
            self,
            trapi_version: Optional[str] = None,
            biolink_version: Optional[str] = None,
            sources: Optional[Dict] = None,
            strict_validation: bool = False
    ):
        """
        :param trapi_version: version of component against which to validate the message (mandatory, no default)
        :type trapi_version: str
        :param biolink_version: Biolink Model (SemVer) release against which the knowledge graph is to be
                                validated (Default: if None, use the Biolink Model Toolkit default version).
        :type biolink_version: Optional[str] = None
        :param sources: Dictionary of validation context identifying the ARA and KP for provenance attribute validation
        :type sources: Dict
        :param strict_validation: if True, some tests validate as 'error'; None or False, simply issue a 'warning'
        :type strict_validation: Optional[bool] = None
        """
        ValidationReporter.__init__(
            self,
            prefix="Validate TRAPI Response",
            trapi_version=trapi_version,
            biolink_version=biolink_version,
            sources=sources,
            strict_validation=strict_validation
        )

    def check_compliance_of_trapi_response(self, message: Dict):
        """
        One stop validation of all components of a TRAPI-schema compliant
        Query Response.Message against a designated Biolink Model release.
        Here, a TRAPI Response message is a Python Dictionary with three entries:

        * Query Graph ("QGraph"): knowledge graph query input parameters
        * Knowledge Graph: output knowledge (sub-)graph containing knowledge (Biolink Model compliant nodes, edges)
                           returned from the target resource (KP, ARA) for the query.
        * Results: a list of (annotated) node and edge bindings pointing into the Knowledge Graph, to represent the
                   specific answers (subgraphs) satisfying the query graph constraints.

        :param message: Response.Message to be validated.
        :type message: Dict

        :returns: Validator cataloging "information", "warning" and "error" messages (may be empty)
        :rtype: ValidationReporter
        """
        if not message:
            self.report("error.trapi.response.message.empty")
        # Sequentially validate the Query Graph, Knowledge Graph then validate
        # the Results (which rely on the validity of the other two components)
        elif self.has_valid_query_graph(message) and \
                self.has_valid_knowledge_graph(message):
            self.has_valid_results(message)

    @staticmethod
    def sample_results(results: List) -> List:
        """

        :param results: List, original list of Results
        :return: List, TEST_DATA_SAMPLE_SIZE sized subset of Results
        """
        sample_size = min(TEST_DATA_SAMPLE_SIZE, len(results))
        result_subsample = results[0:sample_size]
        return result_subsample

    @staticmethod
    def sample_graph(graph: Dict) -> Dict:
        """

        :param graph: Dict, original knowledge graph
        :return: Dict, TEST_DATA_SAMPLE_SIZE sized subset of knowledge graph
        """
        kg_sample: Dict = {
            "nodes": dict(),
            "edges": dict()
        }
        sample_size = min(TEST_DATA_SAMPLE_SIZE, len(graph["edges"]))
        n = 0
        for key, edge in graph['edges'].items():
            kg_sample['edges'][key] = edge
            if 'subject' in edge and edge['subject'] in graph['nodes']:
                kg_sample['nodes'][edge['subject']] = graph['nodes'][edge['subject']]
            if 'object' in edge and edge['object'] in graph['nodes']:
                kg_sample['nodes'][edge['object']] = graph['nodes'][edge['object']]
            n += 1
            if n > sample_size:
                break

        return kg_sample

    def has_valid_query_graph(self, message: Dict) -> bool:
        """
        Validate a TRAPI Query Graph.
        :param message: input message expected to contain the 'query_graph'
        :return: bool, False, if validation errors
        """
        # Query Graph must not be missing...
        if 'query_graph' not in message:
            self.report(code="error.trapi.response.query_graph.missing")
        else:
            # ... nor empty
            query_graph = message['query_graph']
            if not (query_graph and len(query_graph) > 0):
                self.report(code="error.trapi.response.query_graph.empty")
            else:
                # Validate the TRAPI compliance of the Query Graph
                trapi_validator: TRAPISchemaValidator = check_trapi_validity(
                    instance=query_graph,
                    component="QueryGraph",
                    trapi_version=self.trapi_version
                )
                if trapi_validator.has_messages():
                    self.merge(trapi_validator)

                # Validate the Biolink Model compliance of the Query Graph
                biolink_validator = check_biolink_model_compliance_of_query_graph(
                    graph=query_graph,
                    biolink_version=self.biolink_version,
                    # the ValidationReporter calling this function *might*
                    # have an explicit strict_validation override (if not None)
                    strict_validation=self.strict_validation
                )
                if biolink_validator.has_messages():
                    self.merge(biolink_validator)
                    # 'info' and 'warning' messages do
                    # not fully invalidate the query_graph

        # Only 'error' but not 'info' nor 'warning' messages invalidate the overall Message
        return False if self.has_errors() else True

    def has_valid_knowledge_graph(self, message: Dict) -> bool:
        """
        Validate a TRAPI Knowledge Graph.
        :param message: input message expected to contain the 'knowledge_graph'
        :return: bool, False, if validation errors
        """
        # The Knowledge Graph should not be missing
        if 'knowledge_graph' not in message:
            self.report(code="error.trapi.response.knowledge_graph.missing")
        else:
            knowledge_graph = message['knowledge_graph']
            # The Knowledge Graph should also not generally be empty? Issue a warning
            if not (
                    knowledge_graph and len(knowledge_graph) > 0 and
                    "nodes" in knowledge_graph and len(knowledge_graph["nodes"]) > 0 and
                    "edges" in knowledge_graph and len(knowledge_graph["edges"]) > 0
            ):
                self.report(code="warning.response.knowledge_graph.empty")
                # An empty knowledge graph (warning) does not generally invalidate
                # the whole Message, but no more validation tests are needed
            else:
                mapping_validator: MappingValidator = check_node_edge_mappings(knowledge_graph)
                if mapping_validator.has_messages():
                    self.merge(mapping_validator)

                # ...then if not empty, validate a subgraph sample of the associated
                # Knowledge Graph (since some TRAPI response kg's may be huge!)
                kg_sample = self.sample_graph(knowledge_graph)

                # Verify that the sample of the knowledge graph is TRAPI compliant
                trapi_validator: TRAPISchemaValidator = check_trapi_validity(
                    instance=kg_sample,
                    component="KnowledgeGraph",
                    trapi_version=self.trapi_version
                )
                if trapi_validator.has_messages():
                    self.merge(trapi_validator)

                # Verify that the sample of the knowledge graph is
                # compliant to the currently applicable Biolink Model release
                biolink_validator: BiolinkValidator = check_biolink_model_compliance_of_knowledge_graph(
                    graph=kg_sample,
                    biolink_version=self.biolink_version,
                    sources=self.sources,
                    # the ValidationReporter calling this function *might*
                    # have an explicit strict_validation override (if not None)
                    strict_validation=self.strict_validation
                )
                if biolink_validator.has_messages():
                    self.merge(biolink_validator)

        # Only 'error' but not 'info' nor 'warning' messages invalidate the overall Message
        return False if self.has_errors() else True

    def has_valid_results(self, message: Dict) -> bool:
        """
        Validate a TRAPI Results.
        :param message: input message expected to contain the 'results'
        :return: bool, False, if validation errors
        """

        #     :param output_element: test target, as edge 'subject' or 'object'
        #     :type output_element: str
        #     :param output_node_binding: node 'a' or 'b' of the ('one hop') QGraph test query
        #     :type output_node_binding: str

        # The Message.Results key should not be missing?
        if 'results' not in message:
            self.report(code="error.trapi.response.results.missing")
        else:
            results = message['results']
            if not (results and len(results) > 0):
                self.report(code="warning.response.results.empty")
                # An empty result (warning) does not generally invalidate
                # the whole Message, but no more validation tests are needed
            elif not isinstance(results, List):
                # The Message.results should be an array of Result objects
                self.report(code="error.trapi.response.results.non_array")
            else:
                # Validate a subsample of a non-empty Message.results component.
                results_sample = self.sample_results(results)
                for result in results_sample:
                    trapi_validator: TRAPISchemaValidator = check_trapi_validity(
                        instance=result,
                        component="Result",
                        trapi_version=self.trapi_version
                    )
                    if trapi_validator.has_messages():
                        # Record the error messages associated with the Result set then... don't continue
                        self.merge(trapi_validator)

                    # TODO: here, we might wish to compare the Results against the contents of the KnowledgeGraph,
                    #       with respect to node input values from the QueryGraph but this is tricky to do solely
                    #       with the subsamples, which may not completely overlap.

                    # ...Finally, check that the sample Results contained the object of the Query

                    # The 'output_element' is 'subject' or 'object' target (unknown) of retrieval
                    # The 'output_node_binding' is (subject) 'a' or (object) 'b' keys in
                    # the QueryGraph.Nodes to be bound
                    # In principle, we detect which node in the QueryGraph has 'ids' associated with its node record
                    # and assume that the other edge node is the desired target (in the OneHop), so the 'ids'
                    # there should be in the output

                    # object_ids = [r['node_bindings'][output_node_binding][0]['id'] for r in results_sample]
                    # if case[output_element] not in object_ids:
                    #     # The 'get_aliases' method uses the Translator NodeNormalizer to check if any of
                    #     # the aliases of the case[output_element] identifier are in the object_ids list
                    #     output_aliases = get_aliases(case[output_element])
                    #     if not any([alias == object_id for alias in output_aliases for object_id in object_ids]):
                    #         validator.report(
                    #             code=error.results.missing_bindings,
                    #             input_id=case[output_element],
                    #             output_node_binding=output_node_binding
                    #         )
                    #         # data_dump=f"Resolved aliases:\n{','.join(output_aliases)}\n" +
                    #         #           f"Result object IDs:\n{_output(object_ids,flat=True)}"

        # Only 'error' but not 'info' nor 'warning' messages invalidate the overall Message
        return False if self.has_errors() else True



"""Utilities."""
import re
from typing import NamedTuple, Optional
from os import environ
import requests
try:
    from yaml import CLoader as Loader
except ImportError:
    from yaml import Loader


# Undocumented possible local environmental variable
# override of the ReasonerAPI schema access endpoint
GIT_ORG = environ.setdefault('GIT_ORGANIZATION', "NCATSTranslator")
GIT_REPO = environ.setdefault('GIT_REPOSITORY', "ReasonerAPI")

response = requests.get(f"https://api.github.com/repos/{GIT_ORG}/{GIT_REPO}/releases")
releases = response.json()
versions = [
    release["tag_name"][1:]
    for release in releases
    if release["tag_name"].startswith("v")
]


semver_pattern = re.compile(
    r"^(?P<major>0|[1-9]\d*)(\.(?P<minor>0|[1-9]\d*)(\.(?P<patch>0|[1-9]\d*))?)?" +
    r"(?:-(?P<prerelease>(?:0|[1-9]\d*|\d*[a-zA-Z-][\da-zA-Z-]*)(?:\.(?:0|[1-9]\d*|\d*[a-zA-Z-][\da-zA-Z-]*))*))?"
    r"(?:\+(?P<buildmetadata>[\da-zA-Z-]+(?:\.[\da-zA-Z-]+)*))?$"
)


class SemVerError(Exception):
    """Invalid semantic version."""


class SemVerUnderspecified(SemVerError):
    """Semantic version underspecified."""


class SemVer(NamedTuple):
    major: int
    minor: int
    patch: int
    prerelease: Optional[str] = None
    buildmetadata: Optional[str] = None

    @classmethod
    def from_string(cls, string: str, ignore_prefix: Optional[str] = None):
        """
        Initializes a SemVer from a string.

        :param string: str, string encoding the SemVer.
        :param ignore_prefix: Optional[str], if set, gives a prefix of the SemVer string to be ignored before validating
                              the SemVer string value, e.g. a Git Release 'v' character (i.e. v1.2.3); Default: None.
        :return:
        """
        if ignore_prefix:
            string = string.replace(ignore_prefix, "")

        match = semver_pattern.fullmatch(string)

        if match is None:
            raise SemVerError(f"'{string}' is not a valid release version")

        captured = match.groupdict()
        if not all([group in captured for group in ['major', 'minor', 'patch']]):
            raise SemVerUnderspecified(f"'{string}' is missing minor and/or patch versions")
        try:
            return cls(
                *[int(captured[group]) for group in ['major', 'minor', 'patch']],
                *[captured[group] for group in ['prerelease', 'buildmetadata']],
            )
        except TypeError:
            raise SemVerUnderspecified(f"'{string}' is missing minor and/or patch versions")

    def __str__(self):
        """Generate string."""
        value = f"{self.major}"
        if self.minor is not None:
            value += f".{self.minor}"
        if self.patch is not None:
            value += f".{self.patch}"
        if self.prerelease is not None:
            value += f"-{self.prerelease}"
        if self.buildmetadata is not None:
            value += f"+{self.buildmetadata}"
        return value


###########################################
# Deferred SemVer method creation to work #
# around SemVer forward definitions issue #
###########################################
def _semver_ge_(obj: SemVer, other: SemVer) -> bool:
    # Clearcut cases of 'major' release ordering
    if obj.major > other.major:
        return True
    elif obj.major < other.major:
        return False

    # obj.major == other.major
    # Check 'minor' level
    elif obj.minor > other.minor:
        return True
    elif obj.minor < other.minor:
        return False

    # obj.minor == other.minor
    # Check 'patch' level
    elif obj.patch >= other.patch:
        return True
    else:
        # obj.patch < other.patch
        return False


SemVer.__ge__ = _semver_ge_


latest_patch = dict()
latest_minor = dict()
latest_prerelease = dict()
latest = dict()
for version in versions:
    try:
        major, minor, patch, prerelease, buildmetadata = SemVer.from_string(version)
    except SemVerError as err:
        print("\nWARNING:", err)
        continue

    latest_minor[major] = max(minor, latest_minor.get(major, -1))
    latest_patch[(major, minor)] = max(patch, latest_patch.get((major, minor), -1))
    latest_prerelease[(major, minor, patch)] = prerelease \
        if prerelease and not latest_prerelease.get((major, minor, patch), None) else None

    latest[f"{major}"] = str(SemVer(
        major,
        latest_minor[major],
        latest_patch[(major, latest_minor[major])],
    ))
    latest[f"{major}.{minor}"] = str(SemVer(
        major,
        minor,
        latest_patch[(major, minor)],
    ))
    latest[f"{major}.{minor}.{patch}"] = str(SemVer(
        major,
        minor,
        patch,
    ))
    if prerelease:
        latest[f"{major}.{minor}.{patch}-{prerelease}"] = str(SemVer(
            major,
            minor,
            patch,
            prerelease,
            buildmetadata
        ))
    # TODO: we won't bother with buildmetadata for now



"""
General SRI support methods are placed here, e.g. Node Normalizer service access?
"""
import logging
import re
from functools import lru_cache
from json import JSONDecodeError
from typing import Dict, List, Optional

import requests
from kgx.prefix_manager import PrefixManager

logger = logging.getLogger(__name__)

# Endpoint for the Translator Normalizer service
NODE_NORMALIZER_URL = 'https://nodenormalization-sri.renci.org/get_normalized_nodes'


# TODO: not sure to what maxsize for LRU cache
#       should be set so we just take the default
@lru_cache()
def get_aliases(identifier: str):
    """
    Get clique of related identifiers from the Node Normalizer
    """
    if not identifier:
        raise RuntimeError("get_aliases(): empty input identifier?")

    #
    # TODO: maybe check for IRI's here and attempt to translate
    #       (Q: now do we access the prefix map from BMT to do this?)
    # if PrefixManager.is_iri(identifier):
    #     identifier = PrefixManager.contract(identifier)

    # We won't raise a RuntimeError for other various
    # erroneous runtime conditions but simply report warnings
    if not PrefixManager.is_curie(identifier):

        logging.warning(f"get_aliases(): identifier '{identifier}' is not a CURIE thus cannot resolve its aliases?")
        return list()

    # Use the Translator Node Normalizer service to resolve the identifier clique
    response = requests.get(NODE_NORMALIZER_URL, params={'curie': identifier})

    if response.status_code != 200:
        logging.warning(f"get_aliases(): unsuccessful Node Normalizer HTTP call, status code: {response.status_code}")
        return list()
    else:
        try:
            result: Dict = response.json()
        except (JSONDecodeError, UnicodeDecodeError) as je:
            logging.warning(f"get_aliases(): Node Normalizer response JSON could not be decoded: {str(je)}?")
            return list()

    if result and identifier not in result.keys():
        logging.warning(f"get_aliases(): Node Normalizer didn't return the identifier '{identifier}' clique?")
        return list()

    clique = result[identifier]

    aliases: List[str] = list()
    if clique:
        if "id" in clique.keys():
            # TODO: Don't need the canonical identifier for method
            #       but when you do, this is how you'll get it?
            # clique_id = clique["id"]
            # preferred_curie = preferred_id["identifier"]
            # preferred_name = preferred_id["label"]
            if "equivalent_identifiers" in clique.keys():
                aliases: List[str] = [entry["identifier"] for entry in clique["equivalent_identifiers"]]
                aliases.remove(identifier)
                # print(dumps(aliases, indent=2))
            else:
                logging.warning(f"get_aliases(): missing the 'equivalent identifiers' for the '{identifier}' clique?")
        else:
            logging.warning(f"get_aliases(): missing the preferred 'id' for the '{identifier}' clique?")
    else:
        logging.warning(f"get_aliases(): '{identifier}' is a singleton in its clique thus has no aliases...")

    return aliases


CURIE_PATTERN = re.compile(r"^[^ <()>:]*:[^/ :]+$")


def is_curie(s: str) -> bool:
    """
    Check if a given string is a CURIE.

    :param s: str, string to be validated as a CURIE
    :return: bool, whether or not the given string is a CURIE
    """
    # Method copied from kgx.prefix_manager.PrefixManager...
    if isinstance(s, str):
        m = CURIE_PATTERN.match(s)
        return bool(m)
    else:
        return False


def get_reference(curie: str) -> Optional[str]:
    """
    Get the object_id reference of a given CURIE.

    Parameters
    ----------
    curie: str
        The CURIE

    Returns
    -------
    Optional[str]
        The reference of a CURIE

    """
    # Method adapted from kgx.prefix_manager.PrefixManager...
    reference: Optional[str] = None
    if is_curie(curie):
        reference = curie.split(":", 1)[1]
    return reference



"""Setup file for reasoner-validator package."""
from setuptools import setup

with open("README.md", "r") as stream:
    long_description = stream.read()

NAME = 'reasoner-validator'

# Significant API change from release 3.0.0 with respect to validation messaging
# Release 3.1.* targets Biolink Model 3.0.3 or later for optimal validation.
# Validation of Biolink Model releases < 3.0.5 may report a few spurious validation errors.
# This is due to non-backward compatible behaviour of the
# Biolink Model Toolkit, not the validation software per say.
VERSION = '3.1.4'

DESCRIPTION = 'Validation tools for Reasoner API'
URL = 'https://github.com/NCATSTranslator/reasoner-validator'

# Patrick Wang, project creator and emeritus Translator scientist
AUTHOR = 'Richard Bruskiewich, Patrick Wang'

EMAIL = 'richard.bruskiewich@delphinai.com'
REQUIRES_PYTHON = '>=3.8'
LICENSE = 'MIT'

with open("requirements.txt", "r") as FH:
    REQUIREMENTS = FH.readlines()

setup(
    name=NAME,
    author=AUTHOR,
    version=VERSION,
    author_email=EMAIL,
    python_requires=REQUIRES_PYTHON,
    url=URL,
    description="Validation tools for Reasoner API",
    long_description=long_description,
    long_description_content_type="text/markdown",
    license="MIT",
    packages=[
        "reasoner_validator",
        "reasoner_validator.biolink",
        "reasoner_validator.sri",
        "reasoner_validator.trapi"
    ],
    package_data={},
    include_package_data=True,
    install_requires=[r for r in REQUIREMENTS if not r.startswith("#")],
    classifiers=[
        "Programming Language :: Python :: 3.8",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Intended Audience :: Science/Research",
        "Topic :: Scientific/Engineering :: Bio-Informatics"
    ],
    zip_safe=False
)
